--################################################################
--[[
This project will load an animated mesh of a straight wire with a point on one end and loop on the other,
then position the armature to shape it into a fish hook.

--]]

local sw,sh=application:getContentWidth(),application:getContentHeight()
local view=D3.View.new(sw,sh,45,0.1,1000)
local scene=view:getScene() -- we will add all 3D models to it to build our 3D "scene"
application:setOrientation(Application.LANDSCAPE_LEFT)
application:setBackgroundColor(0x00eeee)
local view_container = Sprite.new()
view_container:setPosition(1, 1)
view_container:addChild(view)
stage:addChild(view_container)
Lighting.setLight(5,10,10,0.3)
Lighting.setLightTarget(0,2,0,30,45)

params =
{
	shank_length = 0.9,
	shank_segs = 1,
	initial_angle = 0,
	eye_angle = 55,
	bend_roundness = 1,
}

function updateShankPositions()
	local tx, ty, tz = 0, 0, 0
	local trx, try, trz = 270 - params.eye_angle, 0, 0
	for i = 1, 15 do
		local bone = "s" .. i
		local x, y, z = global_hook.bones[bone]:localToGlobal(0, 0, 0)
		local rx, ry, rz = global_hook.bones[bone]:getRotationX(), global_hook.bones[bone]:getRotationY(), global_hook.bones[bone]:getRotation()
		local sx, sy, sz = global_hook.bones[bone]:getScaleX(), global_hook.bones[bone]:getScaleY(), global_hook.bones[bone]:getScaleZ()
		tx, ty, tz = tx + x, ty + y, tz + z
		global_hook.bones["s" .. i].trot = {trx, try, trz}
		trx = trx + rx
 		shank_pos[i][1] = x shank_pos[i][2] = y shank_pos[i][3] = z
 		shank_pos[i][4][1] = trx shank_pos[i][4][2] = try shank_pos[i][4][3] = trz
		shank_pos[i][5] = { sx, sy, sz }
	end
end


function poseHookBones()
	local shank_length = params.shank_length
	local shank_segs = params.shank_segs
	local initial_angle = params.initial_angle
	local eye_angle = params.eye_angle
	local bend_roundness = params.bend_roundness
	local total_segs = 15
	local curve_segs = total_segs - shank_segs
		
	global_hook.bones.eye:setRotationX(-eye_angle + 180)
	local total_turn = 0
	global_hook.bones.s1:setRotationX(eye_angle - initial_angle)
	total_turn = -initial_angle
	if shank_segs ~= 1 then
		local down_per_seg = 2 * initial_angle / (shank_segs-1)
		for i = 2, shank_segs do
			global_hook.bones["s"..i]:setRotationX(down_per_seg)
			total_turn += down_per_seg
		end
	end
	for i = 2, shank_segs + 1 do
		local x,y,z = global_hook.bones["s"..i]:getPosition()
		global_hook.bones["s"..i]:setPosition(0, 10 * shank_length / shank_segs,0)
	end
	local degrees_to_go = 180 - total_turn
	segs_to_go = total_segs - (shank_segs)
	local min_degr = degrees_to_go / segs_to_go * (bend_roundness + 1) / 2
	for i = shank_segs + 1, total_segs do
		local rel = (i - (shank_segs + 1)) / (total_segs - (shank_segs + 1))
		local segs_to_go = total_segs - 1
		local degr = math.min(degrees_to_go, math.max(min_degr, degrees_to_go * rel * rel * rel))
		degrees_to_go -= degr
		global_hook.bones["s"..i]:setRotationX(degr)
	end
	local x, y, z = global_hook.bones["s" .. total_segs]:getPosition(45)
	global_hook.bones["s" .. total_segs]:setPosition(x, y - 0.75 * bend_roundness, z)
end


local shank_segs = params.shank_segs
	
local size = 48
local rt = RenderTarget.new(size, size)
rt:clear(0x909090)

local prt = RenderTarget.new(1, 1)
prt:clear(0xf0f0f0)
local pbmp = Bitmap.new(prt)
pbmp:setAlpha(0.2)
pbmp:setScaleX(size)

for i = 0, size do
	local a = 1 - (math.abs(i - size / 2) / (size / 2) * math.abs(i - size / 2) / (size / 2))
	pbmp:setAlpha(a * 0.3)
	pbmp:setPosition(0, i)
	rt:draw(pbmp)
end
	
rt:save("|D|hook.png")
global_hook = buildGdx("models/hook_dynamic.json",
	{
		["hookmaterial"]=
		{
			textureFile="|D|hook.png"
		}
	}
)

global_hook_scale = 0.005
global_hook:setScale(global_hook_scale, global_hook_scale, global_hook_scale)
global_hook:setColorTransform(0.75, 0.6, 0.4)
poseHookBones()


local fly_container = Sprite.new()
local room = Sprite.new()
scene:addChild(global_hook)
global_hook:setPosition(0, 3, 0)

stage:addEventListener(Event.ENTER_FRAME,
	function()
		ang = ((ang or 0) + 1) % 360
		d = 20
		view:lookAt(
			d * math.sin(math.rad(ang)), 1.5, d * math.cos(math.rad(ang)),
			0, 1.2, 0,
			0, 1, 0)
		D3Anim.tick()
		
	end
)
